"""DeepSeek generation module."""

from __future__ import annotations

import warnings

import torch

from transformers import AutoModelForCausalLM, AutoTokenizer, GenerationConfig
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class DeepSeekGen(GenerationBase):
    """DeepSeek Generation class."""

    default_model_name: str = 'deepseek-ai/deepseek-llm-7b-chat'
    default_temperature: float = 0.5
    default_output_max_length: int = 500
    default_api_params = {  # noqa: RUF012
        'top_p': 0.9,
        'num_return_sequences': 1,
    }

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        if not self.model_name.startswith('deepseek-ai/'):
            raise Exception(
                f'The given model name {self.model_name} is not provided '
                'by DeepSeek.'
            )

        if self.structured_output:
            warnings.warn(
                'Structured output is not supported yet in '
                f'{self.__class__.__name__}.'
            )

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)

        device_map = 'auto' if self.device_name == 'cuda' else None

        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_name,
            torch_dtype=torch.bfloat16
            if self.device_name == 'cuda'
            else torch.float32,
            device_map=device_map,
        )

        self.model.generation_config = GenerationConfig.from_pretrained(
            self.model_name
        )
        self.model.generation_config.pad_token_id = (
            self.model.generation_config.eos_token_id
        )

    def generate(self, query: str, context: list[str]) -> str:
        """Generate text using DeepSeek model with chat template."""
        messages = [
            {
                'role': 'user',
                'content': f'{query}\nContext: {" ".join(context)}',
            }
        ]

        input_tensor = self.tokenizer.apply_chat_template(
            messages, add_generation_prompt=True, return_tensors='pt'
        ).to(self.model.device)

        model_params = dict(
            max_new_tokens=self.output_max_length,
            do_sample=True,
            temperature=self.temperature,
        )

        self.logs['model_params'] = model_params

        outputs = self.model.generate(input_tensor, **model_params)

        answer: str = str(
            self.tokenizer.decode(
                outputs[0][input_tensor.shape[1] :], skip_special_tokens=True
            )
        )

        return answer.strip()



"""Classes for augmentation with SpaCy embeddings."""

from __future__ import annotations

from hashlib import sha256
from typing import List, cast

import numpy as np
import spacy

from typeguard import typechecked

from rago.augmented.base import AugmentedBase, EmbeddingType


@typechecked
class SpaCyAug(AugmentedBase):
    """Class for augmentation with SpaCy embeddings."""

    default_model_name = 'en_core_web_md'
    default_top_k = 3

    def _setup(self) -> None:
        """Set up the object with initial parameters."""
        self.model = spacy.load(self.model_name)

    def get_embedding(self, content: List[str]) -> EmbeddingType:
        """Retrieve the embedding for a given text using SpaCy."""
        cache_key = sha256(''.join(content).encode('utf-8')).hexdigest()
        cached = self._get_cache(cache_key)
        if cached is not None:
            return cast(EmbeddingType, cached)

        model = cast(spacy.language.Language, self.model)
        embeddings = []

        for text in content:
            doc = model(text)

            # Ensure the model has proper vectors
            if not doc.has_vector:
                raise ValueError(f"Text: '{text}' has no valid word vectors!")

            embeddings.append(doc.vector)

        result = np.array(embeddings, dtype=np.float32)

        # Ensure 2D shape (num_texts, embedding_dim)
        if result.ndim == 1:
            result = result.reshape(1, -1)

        self._save_cache(cache_key, result)
        return result

    def search(
        self, query: str, documents: list[str], top_k: int = 0
    ) -> list[str]:
        """Search an encoded query into vector database."""
        if not hasattr(self, 'db') or not self.db:
            raise Exception('Vector database (db) is not initialized.')

        # Encode the documents and query
        document_encoded = self.get_embedding(documents)
        query_encoded = self.get_embedding([query])
        top_k = top_k or self.top_k or self.default_top_k or 1

        self.db.embed(document_encoded)
        scores, indices = self.db.search(query_encoded, top_k=top_k)

        self.logs['indices'] = indices
        self.logs['scores'] = scores
        self.logs['search_params'] = {
            'query_encoded': query_encoded,
            'top_k': top_k,
        }

        retrieved_docs = [documents[i] for i in indices if i >= 0]

        return retrieved_docs



"""CohereGen class for text generation using Cohere's API."""

from __future__ import annotations

import json

from typing import cast

import cohere
import instructor

from pydantic import BaseModel
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class CohereGen(GenerationBase):
    """Cohere generation model for text generation."""

    default_model_name: str = 'command-r-plus-08-2024'
    default_api_params = {  # noqa: RUF012
        'p': 0.9,
    }

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        model = cohere.ClientV2(api_key=self.api_key)
        self.model = (
            instructor.from_cohere(
                client=model,
                mode=instructor.Mode.COHERE_JSON_SCHEMA,
                model_name=self.model_name,
            )
            if self.structured_output
            else model
        )

    def generate(self, query: str, context: list[str]) -> str | BaseModel:
        """Generate text using Cohere's API."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )
        api_params = self.api_params or self.default_api_params

        if self.structured_output:
            messages = []
            # Explicit instruction to generate JSON output.
            system_instruction = (
                'Generate a JSON object that strictly follows the provided  '
                'JSON schema. Do not include any additional text.'
            )
            if self.system_message:
                system_instruction += ' ' + self.system_message
            messages.append({'role': 'system', 'content': system_instruction})
            messages.append({'role': 'user', 'content': input_text})

            response_format_config = {
                'type': 'json_object',
                'json_schema': (
                    self.structured_output
                    if isinstance(self.structured_output, dict)
                    else self.structured_output.model_json_schema()
                ),
            }
            model_params = {
                'messages': messages,
                'max_tokens': self.output_max_length,
                'temperature': self.temperature,
                'model': self.model_name,
                'response_format': response_format_config,
                **api_params,
            }

            response = self.model.client.chat(**model_params)
            self.logs['model_params'] = model_params
            json_text = response.message.content[0].text
            parsed_dict = json.loads(json_text)
            parsed_model = self.structured_output(**parsed_dict)
            return parsed_model

        if self.system_message:
            messages = [
                {'role': 'system', 'content': self.system_message},
                {'role': 'user', 'content': input_text},
            ]
            model_params = {
                'model': self.model_name,
                'messages': messages,
                'max_tokens': self.output_max_length,
                'temperature': self.temperature,
                **api_params,
            }
            response = self.model.chat(**model_params)
            self.logs['model_params'] = model_params
            return cast(str, response.text)

        model_params = {
            'model': self.model_name,
            'prompt': input_text,
            'max_tokens': self.output_max_length,
            'temperature': self.temperature,
            **api_params,
        }
        response = self.model.generate(**model_params)
        self.logs['model_params'] = model_params
        return cast(str, response.generations[0].text.strip())
